package controller;

import java.util.LinkedList;
import java.util.concurrent.BlockingQueue;

import champ2010client.Action;
import champ2010client.SensorModel;

/**
 * Trieda vytvarajuca zaznamy o prejdenych usekoch trate. Bezi pri tom v inom vlakne
 * ako ovladaci modul vozidla.
 * 
 * @author Ivan Valencik
 */
public class TurnDetector implements Runnable
{
	public enum Direction { STRAIGHT, RIGHT, LEFT };
	
	private static final int TURN_SENSITIVITY = 4;

	private final BlockingQueue<TickInfo> tickInfoPipe;
	
	private Record record = null;
	private Record builtSector = null;
	private Record previousSector = null;
	
	// so zoznamom je povolene manipulovat, len ak je cinnost vzhladom nan synchronizovana
	public final LinkedList<Record> trackSectors = new LinkedList<Record>();

	public TurnDetector(BlockingQueue<TickInfo> source)
	{
		tickInfoPipe = source;
	}

	public void run()
	{
		try {
			while (true)
			{
				TickInfo ti = tickInfoPipe.take();
				inform(ti.getAction(), ti.getSensorModel());
//				if (tickInfoPipe.size() > 0)
//					System.out.println("V rade caka: " + tickInfoPipe.size() + " udajov.");
			}
		} catch (InterruptedException e) {
			System.err.println("Error: chyba pri prijmani akcie do modelu trate.");
			e.printStackTrace();
		}
	}

	/**
	 * Na zaklade akcie urci, ci doplni vytvarany zaznam alebo zacne pracovat s novym.
	 * @param action vykonana akcia
	 */
	public void inform(Action action, SensorModel sm)
	{
		if (record == null)
		{
			record = new Record(action.steering, sm.getDistanceFromStartLine(),
					sm.getSpeed());
			return;
		}

		switch (record.getDirection())
		{
			case RIGHT:
				if (action.steering < 0)
					record.update(action.steering, sm.getSpeed());
				else
				{
					submitRecord(record);
					record = new Record(action.steering, sm.getDistanceFromStartLine(),
							sm.getSpeed());
				}
				break;
			case LEFT:
				if (action.steering > 0)
					record.update(action.steering, sm.getSpeed());
				else
				{
					submitRecord(record);
					record = new Record(action.steering, sm.getDistanceFromStartLine(),
							sm.getSpeed());				}
				break;
			case STRAIGHT:
				if (action.steering == 0)
					record.update(action.steering, sm.getSpeed());
				else
				{
					submitRecord(record);
					record = new Record(action.steering, sm.getDistanceFromStartLine(),
							sm.getSpeed());				}
				break;
		}
	}

	/**
	 * Spajanie usekov zakruty a vytvaranie zoznamu zakrut.
	 * @param record zaznam o useku zakruty
	 */
	private void submitRecord(Record record)
	{
		Direction dir = null;
		
		if (builtSector == null)
		{
			builtSector = record;
			return;
		}

		if (previousSector == null)
		{
			previousSector = record;
			return;
		}

		// ak je pomer medzi usekmi nad urcenou hodnotou, tak vyberie typ dlhsieho useku
		if (previousSector.getDuration() / record.getDuration() > TURN_SENSITIVITY)
			dir = previousSector.getDirection();
		if (record.getDuration() / previousSector.getDuration() > TURN_SENSITIVITY)
			dir = record.getDirection();

		/*
		ak isiel rovno a ide rovno, tak rovno
		ak isiel rovno a zataca, tak zatoci
		ak zatacal a zataca rovnakym smerom, tak v tom pokracuje
		ak zatacal a zataca opacnym smerom, tak chce predposledny zaznam
		*/
		if (dir == null)
			if (previousSector.getDirection() == Direction.STRAIGHT)
				if (record.getDirection() == Direction.STRAIGHT)
					dir = Direction.STRAIGHT;
				else
					dir = record.getDirection();
			else
				dir = previousSector.getDirection();
		
		if (dir == builtSector.getDirection())
			builtSector.merge(previousSector);
		else
		{
			synchronized (trackSectors)
			{
				// spaja useky rovnakeho typu, ktore rozoznalo za sebou
				if (!trackSectors.isEmpty() && 
						trackSectors.getLast().getDirection() == builtSector.getDirection())
					trackSectors.getLast().merge(builtSector);
				else
				{
					if (!trackSectors.isEmpty())
						Util.printRecord(trackSectors.getLast());
					trackSectors.addLast(builtSector);
				}
			}
			builtSector = previousSector;
		}

		previousSector = record;
	}
}
